__author__ = "P.Mohr"
__copyright__ = "26.03.2020, P.Mohr"
__version__ = "1.0"
__email__ = "philipp.mohr@tuhh.de"
__doc__ = """This script includes helpful functions to extended PyOpenCl functionality."""

import os
import re
import time
from abc import abstractmethod, ABC
from dataclasses import dataclass, field
from enum import Enum
from functools import lru_cache
from pathlib import Path
from typing import Union, Tuple, List, Callable

import numpy as np
import numpy.typing as npt
import pyastyle
from mako import exceptions
from mako.template import Template
import pyopencl as cl
from pyopencl._cl import Device
from pyopencl.array import Array as ClArray

from pyopencl_extension import CommandQueue, Array, to_device
from pyopencl_extension.helpers.general import write_string_to_file
from pyopencl_extension.modifications_pyopencl.command_queue import QueueProperties, get_current_queue
from pyopencl_extension.modifications_pyopencl.context import Context, get_devices
from pyopencl_extension.types.auto_gen.cl_types import ClTypesScalar
from pyopencl_extension.types.utilities_np_cl import c_name_from_dtype, scalar_type_from_vec_type, \
    get_vec_size, Types, number_vec_elements_of_cl_type, VEC_INDICES
from pyopencl_extension.emulation import create_py_file_and_load_module, unparse_c_code_to_python
import logging


@dataclass
class LocalArray:
    shape: int
    dtype: npt.DTypeLike
    cl_local_memory: cl.LocalMemory = field(init=False, default=None)

    def __post_init__(self):
        self.cl_local_memory = cl.LocalMemory(int(self.shape * np.dtype(self.dtype).itemsize))


TypesClArray = Union[Array, ClArray]
TypesDefines = Union[str, float, int, bool]
TypesReplacement = Union[str, float, int, bool]
TypesArgArrays = Union[np.ndarray, Array, ClArray, LocalArray]
_ = ClTypesScalar
TypesArgScalar = Union[int, float,
                       _.char, _.short, _.int, _.long, _.uchar, _.ushort, _.uint, _.ulong, _.half, _.float, _.double]
# TypesKernelArg = Union[Array, TypesDefines] # todo: remove?

preamble_activate_double = """
#if defined(cl_khr_fp64)  // Khronos extension available?
#pragma OPENCL EXTENSION cl_khr_fp64 : enable
#define PYOPENCL_DEFINE_CDOUBLE
#elif defined(cl_amd_fp64)  // AMD extension available?
#pragma OPENCL EXTENSION cl_amd_fp64 : enable
#define PYOPENCL_DEFINE_CDOUBLE
#endif
"""

preamble_activate_complex_numbers = """
#include <pyopencl-complex.h>
#define TP_ROOT ${cplx_type}
"""


def preamble_precision(precision: str = 'single'):
    """
    This function generates preamble to support either single or double precision floating point numbers.
    :param precision:
    :return:
    """
    if precision == 'single':
        return """
#define PI 3.14159265359f
            """
    elif precision == 'double':
        return preamble_activate_double + """\n
#define PI 3.14159265358979323846
            """
    else:
        raise NotImplementedError()


def preamble_generic_type_operations(number_format: str = 'real', precision: str = 'single'):
    """
    This function returns a preamble which defines how generic operations are executed on device.

    This becomes especially important when dealing with complex types which  OpenCl does not support out of the box.
    As a solution, pyopencl-complex.h includes several basic function for complex operations.

    E.g. consider a kernel which adds two numbers, however the input type can be real or complex valued.
    Typically one would implement c = a + b. However, OpenCl does not support + operation when a and b are complex
    valued. Therefore using this preamble one can write c = ADD(a,b). ADD acts a a generic operation which supports
    real and complex input depending on selection for number_format.


    :param number_format: 'real' or 'complex
    :param precision: 'single' or 'double'
    :return: preamble to support generic operations
    """

    if number_format == 'complex':
        cplx_type_pyopencl = {'single': 'cfloat',
                              'double': 'cdouble'}[precision]
        return preamble_activate_complex_numbers + """
#define MUL ${cplx_type}_mul
#define ADD ${cplx_type}_add
#define SUB ${cplx_type}_sub
#define ABS ${cplx_type}_abs
#define RMUL ${cplx_type}_rmul
#define NEW ${cplx_type}_new
#define CONJ ${cplx_type}_conj
#define REAL(x) x.real
#define IMAG(x) x.imag
    """.replace('${cplx_type}', cplx_type_pyopencl)
    elif number_format == 'real':
        return """
#define MUL(x,y) (x*y)
#define ADD(x,y) (x+y)
#define SUB(x,y) (x-y)
#define ABS(x) (fabs(x))
#define RMUL(x,y) (x*y)
#define NEW(x,y) (x)
#define CONJ(x) (x)
#define REAL(x) (x)
#define IMAG(x) (0)
        """
    else:
        raise NotImplementedError()


def catch_invalid_argument_name(name: str):
    """
    E.g. when using certain argument names like 'channel' the opencl compiler throws a compilation error, probably
    because channel is an reserved opencl command. There we replace those names by appending '_' character.

    :param name:
    :return:
    """
    invalid_names = ['channel']
    if name in invalid_names:
        raise ValueError('Invalid opencl name: \'{}\' used.'.format(name))
    else:
        return name


class OrderInMemory(Enum):
    C_CONTIGUOUS: str = 'c_contiguous'
    F_CONTIGUOUS: str = 'f_contiguous'


@dataclass
class ArgBase(ABC):
    # too much restriction, shape of array might change during runtime
    # shape: Tuple[int, ...] = (1,)  # default: argument is scalar

    @property
    @abstractmethod
    def address_space_qualifier(self) -> str:
        # __global, __local, __private, __constant
        pass

    @property
    @abstractmethod
    def dtype(self) -> npt.DTypeLike:
        # __global, __local, __private, __constant
        pass

    def to_string(self, name):
        new_name = catch_invalid_argument_name(name)
        if type(self) in [Scalar]:  # scalar
            return '{} {} {}'.format(self.address_space_qualifier, c_name_from_dtype(self.dtype), new_name)
        else:  # array
            return '{} {} *{}'.format(self.address_space_qualifier, c_name_from_dtype(self.dtype), new_name)


@dataclass
class Scalar(ArgBase):
    dtype: npt.DTypeLike = field(default=Types.int)
    address_space_qualifier: str = field(default='')
    default: npt.DTypeLike = field(init=False, default=None)

    def __post_init__(self):
        if np.isscalar(self.dtype):
            self.default = self.dtype
            if type(self.default) == float:
                self.dtype = Types.double
            elif type(self.default) == int:
                self.dtype = Types.int
            else:
                self.dtype = type(self.dtype)


@dataclass
class Pointer(ArgBase, ABC):
    dtype: npt.DTypeLike = field(default=Types.int)
    address_space_qualifier: str = field(init=False, default='__global')


@dataclass
class Private(Pointer):
    address_space_qualifier: str = field(init=False, default='__private')


@dataclass
class Local(Pointer):
    dtype: npt.DTypeLike | LocalArray = field(default=Types.int)
    address_space_qualifier: str = field(init=False, default='__local')
    order_in_memory: OrderInMemory = OrderInMemory.C_CONTIGUOUS
    default: cl.LocalMemory = field(init=False, default=None)

    def __post_init__(self):
        if isinstance(self.dtype, LocalArray):
            self.default = self.dtype.cl_local_memory
            self.dtype = self.dtype.dtype


@dataclass
class Global(Pointer):
    dtype: npt.DTypeLike | TypesClArray = field(default=Types.int)
    read_only: bool = False  # adds 'const' qualifier to let compiler know that global array is never written
    order_in_memory: OrderInMemory = OrderInMemory.C_CONTIGUOUS
    address_space_qualifier: str = field(init=False, default='__global')
    default: TypesClArray = field(init=False, default='')

    def __post_init__(self):
        if isinstance(self.dtype, TypesClArray.__args__):
            self.default = self.dtype
            self.dtype = self.dtype.dtype

            if self.read_only:
                self.address_space_qualifier = 'const __global'


@dataclass
class Constant(Pointer):
    """
    const is only a hint for the compiler that the data does not change
    __constant leads to usage of very fast constant cache memory which is shared among
    multiple compute units. From AMD optimization guide, e.g. we can read 4 bytes/cycles.
    Local memory can be ready twice as fast with 8bytes/cycle, however local memory is a even more scarce resource.

    https://stackoverflow.com/questions/17991714/opencl-difference-between-constant-memory-and-const-global-memory/50931783
    """
    dtype: npt.DTypeLike | TypesClArray = field(default=Types.int)
    order_in_memory: str = OrderInMemory.C_CONTIGUOUS
    address_space_qualifier: str = field(init=False, default='__constant')
    default: npt.DTypeLike | TypesClArray = field(init=False, default='')

    def __post_init__(self):
        if isinstance(self.dtype, TypesClArray.__args__):
            self.default = self.dtype
            self.dtype = self.dtype.dtype


def template(func: Union['Kernel', 'Function']) -> str:
    body = ''.join(func.body)
    tpl = func.header + '\n{' + body + '}\n'
    args = [value.to_string(key) + ',' for key, value in func.args.items()]
    args = '{}'.format('\n'.join(args))
    args = args[:-1]  # remove last comma
    replacements = {'name': func.name,
                    'args': args,
                    'returns': c_name_from_dtype(func.returns)}
    for key, value in func.replacements.items():
        replacements[key] = str(value)
    try:  # todo: e.g. if replacement has been forgotten, still save template as file
        tpl = Template(tpl).render(**replacements)
    except:
        raise ValueError(exceptions.text_error_template().render())

    defines = '\n'.join(['#define {} {}'.format(key, str(value)) for key, value in func.defines.items()])
    tpl_formatted = pyastyle.format('{}\n\n{}'.format(defines, tpl), '--style=allman --indent=spaces=4')
    return tpl_formatted


@dataclass
class FunctionBase(ABC):
    name: str = 'func'
    args: dict[str, Union[TypesArgArrays, TypesArgScalar, Scalar, Global, Local, Private, Constant]] = \
        field(default_factory=lambda: {})
    body: Union[List[str], str] = field(default_factory=lambda: [])
    replacements: dict[str, TypesReplacement] = field(default_factory=lambda: {})
    type_defs: dict[str, npt.DTypeLike] = field(default_factory=lambda: {})  # todo
    defines: dict[str, TypesDefines] = field(default_factory=lambda: {})
    functions: List['Function'] = field(default_factory=lambda: [])

    def __post_init__(self):
        if isinstance(self.body, str):
            self.body = [self.body]
        self._prepares_args()

    @property
    def header(self):
        return '${returns} ${name}(${args})'

    @property
    @abstractmethod
    def template(self) -> str:
        pass

    @staticmethod
    def _prepare_arg(v):
        """
        This is a convenience feature. Arguments might be provided only as numpy array, python integer or float etc.
        Therefore, this function adds appropriate pointer type.
        """
        if isinstance(v, np.ndarray):
            g_arg = Global(v.dtype)
            g_arg.default = v
            return g_arg
        elif isinstance(v, TypesClArray.__args__):
            return Global(v)
        elif isinstance(v, LocalArray):
            return Local(v)
        elif isinstance(v, TypesArgScalar.__args__):
            return Scalar(v)
        else:
            return v

    def _prepares_args(self):
        self.args = {k: self._prepare_arg(v) for k, v in self.args.items()}


@dataclass
class Function(FunctionBase):
    @property
    def template(self) -> str:
        return template(self)

    returns: npt.DTypeLike = field(default_factory=lambda: np.dtype(np.void))

    def __str__(self) -> str:
        return super().__str__() + str(self.returns)


KernelGridType = Union[Tuple[int], Tuple[int, int], Tuple[int, int, int]]


class Compilable:
    @abstractmethod
    def compile(self, context: Context = None, emulate: bool = False):
        pass

    @staticmethod
    def get_default_dir_pycl_kernels():
        return Path(os.getcwd()).joinpath('cl_py_modules')


@dataclass
class Kernel(FunctionBase, Compilable):
    def compile(self, context: Context = None, emulate: bool = False, file='$default_path', **kwargs):
        # if kernel is compiled put defines on top of program
        defines = self.defines
        self.defines = {}  # ensures that defines are not placed before kernel another time
        Program(defines=defines, kernels=[self]).compile(context=context, emulate=emulate, file=file)
        self.defines = defines  # reconstructs defines
        return self.callable_kernel

    global_size: KernelGridType = None
    local_size: KernelGridType = None
    returns: npt.DTypeLike = field(default_factory=lambda: np.dtype(np.void), init=False)
    callable_kernel: 'CallableKernel' = field(default_factory=lambda: None, init=False)

    def __str__(self) -> str:
        return super().__str__()

    @property
    def template(self) -> str:
        return template(self)

    @property
    def header(self):
        return '__kernel ${returns} ${name}(${args})'

    def __call__(self, global_size: KernelGridType = None, local_size: KernelGridType = None, **kwargs):
        if self.callable_kernel is None:
            self.compile()  # compile kernel if not compiled
        return self.callable_kernel(global_size=global_size, local_size=local_size, **kwargs)


def _get_all_funcs(f: FunctionBase, flat_list=None) -> List[FunctionBase]:
    if flat_list is None:
        flat_list = []
        for sub_f in f.functions:
            _get_all_funcs(sub_f, flat_list)
        flat_list.append(f)
        return flat_list
    else:
        flat_list.extend([_get_all_funcs(sub_f, flat_list) for sub_f in f.functions])
        flat_list.append(f)


def _get_list_with_unique_functions(functions, kernels):
    functions_in_kernels = [f for k in kernels for f in k.functions]
    all_funcs = [_f for f in functions + functions_in_kernels for _f in _get_all_funcs(f)]
    all_funcs_unique, _func_names = [], []
    for f in all_funcs:
        if len(_func_names) == 0 or f.name not in _func_names:
            all_funcs_unique.append(f)
            _func_names.append(f.name)
    return all_funcs_unique


@dataclass
class Program(Compilable):
    """
    Models an OpenCl Program containing functions or kernels.
    """

    def compile(self, context: Context = None, emulate: bool = False,
                file: str = '$default_path') -> 'ProgramContainer':

        return compile_cl_program(self, context, emulate, file)

    functions: List[Function] = field(default_factory=lambda: [])
    kernels: List[Kernel] = field(default_factory=lambda: [])
    defines: dict[str, TypesDefines] = field(default_factory=lambda: {})
    type_defs: dict[str, npt.DTypeLike] = field(default_factory=lambda: {})

    @staticmethod
    def _arg_to_str_for_hash(name, arg: ArgBase):
        return name + str(type(arg)) + str(hash(arg.dtype)) + arg.address_space_qualifier

    @staticmethod
    def _func_to_str_for_hash(func: FunctionBase):
        str_funcs = ''.join(Program._func_to_str_for_hash(knl) for knl in func.functions)
        str_args = ''.join(Program._arg_to_str_for_hash(k, v) for k, v in func.args.items())
        str_repl = ''.join(k + str(v) for k, v in func.replacements.items())
        str_type_defs = ''.join(k + str(v) for k, v in func.type_defs.items())
        str_defines = ''.join(k + str(v) for k, v in func.defines.items())
        str_body = ''.join(func.body)
        return str_funcs + str_type_defs + str_defines + str_repl + func.header + func.name + str_args + str_body

    def __hash__(self) -> int:
        str_funcs = ''.join(self._func_to_str_for_hash(knl) for knl in self.functions)
        str_kernels = ''.join(self._func_to_str_for_hash(knl) for knl in self.kernels)
        str_type_defs = ''.join(k + str(v) for k, v in self.type_defs.items())
        str_defines = ''.join(k + str(v) for k, v in self.defines.items())
        prog_str = str_defines + str_type_defs + str_funcs + str_kernels
        return int(str(hash(prog_str)) + str(abs(hash(prog_str + 'something'))))  # double hash to avoid collisions

    @property
    def rendered_template(self):
        all_funcs = _get_list_with_unique_functions(self.functions, self.kernels)

        functions = [f.template for f in all_funcs] + [k.template for k in self.kernels]
        functions = '\n'.join(functions)

        if 'double' in functions:
            _preamble_precision = preamble_precision('double')
        else:
            _preamble_precision = preamble_precision('single')

        if 'cfloat_t' in functions:
            _preamble_generic_operations = preamble_generic_type_operations('complex', 'single')
        elif 'cdouble_t' in functions:
            _preamble_generic_operations = preamble_generic_type_operations('complex', 'double')
        else:
            _preamble_generic_operations = preamble_generic_type_operations('real')

        _preamble_wi_ops = """
#define gid0 get_global_id(0)
#define gid1 get_global_id(1)
#define gid2 get_global_id(2)
#define gs0 get_global_size(0)
#define gs1 get_global_size(1)
#define gs2 get_global_size(2)

#define lid0 get_local_id(0)
#define lid1 get_local_id(1)
#define lid2 get_local_id(2)
#define ls0 get_local_size(0)
#define ls1 get_local_size(1)
#define ls2 get_local_size(2)

#define wid0 get_group_id(0)
#define wid1 get_group_id(1)
#define wid2 get_group_id(2)
        """
        preamble_buff_t = f'{_preamble_precision}\n{_preamble_generic_operations}\n{_preamble_wi_ops}'

        # join program typedefs with typedefs from kernels and functions
        # todo: consider replacing type strings directly to avoid name conflicts
        def update_and_checks_for_duplicates_same_type(items: dict, dic: dict):
            for key, value in items.items():
                if key in dic:
                    if not dic[key] == value:
                        raise ValueError('Same type def name for different types')
                else:
                    dic[key] = value

        [update_and_checks_for_duplicates_same_type(func.type_defs, self.type_defs) for func in self.functions]
        [update_and_checks_for_duplicates_same_type(func.type_defs, self.type_defs) for func in self.kernels]

        # remove since defines are inserted before function/kernels
        # [update_and_checks_for_duplicates_same_type(func.defines, self.defines) for func in self.functions]
        # [update_and_checks_for_duplicates_same_type(func.defines, self.defines) for func in self.kernels]

        defines = '\n'.join(['#define {} {}'.format(key, str(value)) for key, value in self.defines.items()])

        type_defs = '\n'.join(
            ['typedef {c_name} {new_name};\n#define convert_{new_name}(x) convert_{c_name}(x)'.format(
                c_name=c_name_from_dtype(value),
                new_name=str(key))
                for key, value in self.type_defs.items()])

        tpl_all = self._get_tpl(preamble_buff_t, defines, type_defs, functions)

        tpl_formatted = pyastyle.format(tpl_all, '--style=allman --indent=spaces=4')
        return tpl_formatted

    def _get_tpl(self, preamble_complex, defines, type_defs, functions):
        return '{}\n\n{}\n\n{}\n\n{}\n\n'.format(preamble_complex, defines, type_defs, functions)


# by default not files are written to reduce disk usage
_create_cl_file = False


def create_cl_files(flag=True):
    global _create_cl_file
    _create_cl_file = flag


def build_for_device(context: Context, template_to_be_compiled: str, file: str = None) -> cl.Program:
    if _create_cl_file and file is not None:
        write_string_to_file(template_to_be_compiled, file + '.cl', b_logging=True)
    try:
        program = cl.Program(context, str(template_to_be_compiled)).build()
    except Exception as error:
        tpl_line_numbers = '\n'.join(
            ['{:<4}{}'.format(i + 1, line) for i, line in enumerate(template_to_be_compiled.split('\n'))])
        raise ValueError('\n{}\n\n{}'.format(tpl_line_numbers, str(error)))
    return program


# Todo: Find good structure for modeling cl and python kernels
@dataclass
class CallableKernel(ABC):
    kernel_model: Kernel

    def __getattr__(self, name):
        if name in self.kernel_model.args.keys():
            return self.kernel_model.args[name].default
        return super().__getattribute__(name)

    @abstractmethod
    def __call__(self, global_size: KernelGridType = None,
                 local_size: KernelGridType = None,
                 **kwargs):
        pass

    @staticmethod
    def _typing_scalar_argument(arg_model: Union[Scalar, Scalar],
                                scalar_value_provided: TypesArgScalar):
        if get_vec_size(arg_model.dtype) == 1:
            return np.dtype(arg_model.dtype).type(scalar_value_provided)
        else:
            dtype_scalar = scalar_type_from_vec_type(arg_model.dtype)
            scalar = np.dtype(dtype_scalar).type(scalar_value_provided)  # converts to bytes like object
            return scalar.astype(arg_model.dtype)  # converts to vector type

    @staticmethod
    def _prepare_arguments(queue: CommandQueue, knl: Kernel, **kwargs):
        global_size = kwargs.pop('global_size', None)
        local_size = kwargs.pop('local_size', None)
        global_size = knl.global_size if global_size is None else global_size
        local_size = knl.local_size if local_size is None else local_size

        supported_kws = [k for k in knl.args.keys()]
        kw_not_in_kernel_arguments = [kw for kw in kwargs if kw not in supported_kws]
        if len(kw_not_in_kernel_arguments) > 0:
            raise ValueError(
                f'keyword argument {kw_not_in_kernel_arguments} does not exist in kernel argument list {supported_kws}')

        # If kernel arguments are of type np.ndarray they are converted to cl arrays here
        # This is done here, since thq queue is available at this point for sure.
        # todo: deal with case if kwarg is numpy argument
        def deal_with_np_arrays(v):
            if isinstance(v, Global) and isinstance(v.default, np.ndarray):
                v.default = to_device(ary=v.default, queue=queue)
                return v
            else:
                return v

        knl.args = {k: deal_with_np_arrays(v) for k, v in knl.args.items()}

        # set default arguments. Looping over kernel model forces correct order of arguments
        args_call = [kwargs.pop(key, value.default if isinstance(value, (Constant, Global, Scalar, Local)) else None)
                     for key, value in knl.args.items()]
        if any(arg is None for arg in args_call):
            raise ValueError('Argument equal to None can lead to system crash')

        if global_size is None:
            raise ValueError('global_size not provided!')
        if global_size == ():
            raise ValueError('Global size is empty')
        if 0 in global_size:
            raise ValueError(f'Parameter in global size {global_size} equal to zero')
        if local_size is not None and 0 in local_size:
            raise ValueError(f'Parameter in local size {local_size} equal to zero')
        # convert scalar argument to correct type. E.g. argument can be python int and is converted to char
        args_model = list(knl.args.values())
        args_call = [CallableKernel._typing_scalar_argument(args_model[i], arg)
                     if type(args_model[i]) in [Scalar, Scalar] else arg
                     for i, arg in enumerate(args_call)]
        # if argument of type LocalArray extract cl.LocalMemory instance to be passed as argument
        args_call = [arg.cl_local_memory if isinstance(arg, LocalArray) else arg for arg in args_call]
        # check if buffer have same type as defined in the kernel function header
        b_types_equal = [args_call[i].dtype == v.dtype for i, v in enumerate(args_model) if isinstance(v, Global)]
        if not np.all(b_types_equal):
            idx_buffer_list = int(np.argmin(b_types_equal))
            idx = [i for i, kv in enumerate(knl.args.items()) if isinstance(kv[1], Global)][idx_buffer_list]
            buffer_name = [k for k, v in knl.args.items()][idx]
            buffer_type_expected = args_model[idx].dtype
            buffer_type_call = args_call[idx].dtype
            raise ValueError(f'Expected buffer argument \'{buffer_name}\' with type {buffer_type_expected} '
                             f'but got buffer with type {buffer_type_call}')

        # check if buffer elements of array arguments have memory order as expected (c or f contiguous)
        def b_array_memory_order_as_expected(ary_model: Global, ary_call: TypesClArray):
            if ary_model.order_in_memory == OrderInMemory.C_CONTIGUOUS:
                return ary_call.flags.c_contiguous
            else:  # f_contiguous
                return ary_call.flags.f_contiguous

        knl_args_invalid_memory_order = [(k, v) for idx, (k, v) in enumerate(knl.args.items())
                                         if isinstance(v, Global) and
                                         not b_array_memory_order_as_expected(v, args_call[idx])]
        if len(knl_args_invalid_memory_order) > 0:
            msg = '\n'.join([f'Array argument \'{arg[0]}\' is not {arg[1].order_in_memory} (as defined in Kernel)'
                             for arg in knl_args_invalid_memory_order])
            raise ValueError(msg)
        non_supported_types = [np.ndarray]
        if any(_ := [type(arg) in non_supported_types for arg in args_call]):
            raise ValueError(f'Type of argument \'{list(knl.args.items())[np.where(_)[0][0]][0]}\' is not supported in '
                             f'kernel call')
        return global_size, local_size, args_call


@dataclass
class CallableKernelEmulation(CallableKernel):
    function: Callable

    def __call__(self,
                 global_size: KernelGridType = None,
                 local_size: KernelGridType = None,
                 **kwargs: Union[TypesClArray, object]) -> cl.Event:
        # e.g. if two kernels of a program shall run concurrently, this can be enable by passing another queue here
        queue = kwargs.pop('queue', get_current_queue())
        global_size, local_size, args = self._prepare_arguments(queue=queue, knl=self.kernel_model,
                                                                global_size=global_size,
                                                                local_size=local_size, **kwargs)
        self.function(global_size, local_size, *args)
        # create user event with context retrieved from first arg of type Array
        event = cl.UserEvent([_ for _ in args if isinstance(_, TypesClArray.__args__)][0].context)
        event.set_status(cl.command_execution_status.COMPLETE)
        return event


@dataclass
class CallableKernelDevice(CallableKernel):
    compiled: cl.Kernel

    @staticmethod
    def check_local_size_not_exceeding_device_limits(device: Device, local_size):
        # E.g. on nvidia the local size might be individually limited to be (1024,1024,64).
        # This shall trigger an exception, when wrong local size is provided.
        if local_size is not None and any([desired_local_size > device.max_work_item_sizes[dim]
                                           for dim, desired_local_size in enumerate(local_size)]):
            raise ValueError(f'Requested local dimensions {local_size} exceed {device.max_work_item_sizes=}')

    def __call__(self,
                 global_size: KernelGridType = None,
                 local_size: KernelGridType = None,
                 **kwargs) -> cl.Event:
        # e.g. if two kernels of a program shall run concurrently, this can be enable by passing another queue here
        queue = kwargs.pop('queue', get_current_queue())
        assert self.compiled.context.int_ptr == queue.context.int_ptr
        global_size, local_size, args = self._prepare_arguments(queue=queue, knl=self.kernel_model,
                                                                global_size=global_size,
                                                                local_size=local_size, **kwargs)
        self.check_local_size_not_exceeding_device_limits(queue.device, local_size)
        # extract buffer from cl arrays separate, since in emulation we need cl arrays
        args_cl = [arg.data if isinstance(arg, TypesClArray.__args__) else arg for i, arg in enumerate(args)]
        event = self.compiled(queue, global_size, local_size, *args_cl)
        queue.add_event(event, self.kernel_model.name)
        return event


@dataclass
class ProgramContainer:
    """
    Responsibility:
    A callable kernel is returned with program.kernel_name. Depending on value of b_run_python_emulation a call of this
    kernel is executed on device or in emulation.
    """
    program_model: Program
    file: str
    init: CommandQueue
    callable_kernels: dict[str, Union[CallableKernelEmulation, CallableKernelDevice]] = None

    def __getattr__(self, name) -> CallableKernel:
        if name in self.callable_kernels:
            return self.callable_kernels[name]
        else:
            return super().__getattribute__(name)


# https://stackoverflow.com/questions/1988804/what-is-memoization-and-how-can-i-use-it-in-python
class MemoizeKernelFunctions:
    def __init__(self, f):
        self.f = f
        self.memo = {}

    def __call__(self, program_model: Program, context: Context, file: str = None):
        # body = ''.join(program_model.rendered_template)
        _id = hash(f'{hash(context)}{hash(program_model)}')
        if _id not in self.memo:
            self.memo[_id] = self.f(program_model, context, file)
        return self.memo[_id]


@MemoizeKernelFunctions
def compile_cl_program_device(program_model: Program, context: Context = None, file: str = None) -> dict[str, Kernel]:
    code_cl = program_model.rendered_template
    program = build_for_device(context, code_cl, file)
    kernels_model = program_model.kernels
    # set scalar arguments for each kernel from kernel model
    callable_kernels = {knl.function_name: knl
                        for i, knl in enumerate(program.all_kernels())}
    for i, knl in enumerate(kernels_model):
        arg_types = [arg.dtype if type(arg) in [Scalar, Scalar] else None
                     for _, arg in kernels_model[i].args.items()]
        callable_kernels[knl.name].set_scalar_arg_dtypes(arg_types)
    return callable_kernels


@MemoizeKernelFunctions
def compile_cl_program_emulation(program_model: Program, context: Context, file: str = None,
                                 *args, **kwargs) -> dict[str, Callable]:
    code_cl = program_model.rendered_template
    write_string_to_file(code_cl, file + '.cl', b_logging=True)  # creates file with cl code
    code_py = unparse_c_code_to_python(code_c=code_cl)
    module = create_py_file_and_load_module(code_py, file)
    kernels_model = program_model.kernels
    callable_kernels = {knl.name: module.__getattribute__(knl.name) for knl in kernels_model}
    return callable_kernels


def compile_cl_program(program_model: Program, context: Context = None, emulate: bool = False,
                       file: str = '$default_path') -> ProgramContainer:
    t_ns_start = time.perf_counter_ns()
    # deal with file name
    if isinstance(file, Path):
        file = str(file)
    if file == '$default_path':
        file = str(program_model.get_default_dir_pycl_kernels().joinpath(program_model.kernels[0].name))

    if context is None:
        context = get_current_queue().context

    dict_kernels_program_model = {knl.name: knl for knl in program_model.kernels}
    if emulate:
        dict_emulation_kernel_functions = compile_cl_program_emulation(program_model, context, file)
        callable_kernels = {k: CallableKernelEmulation(kernel_model=dict_kernels_program_model[k], function=v)
                            for k, v in dict_emulation_kernel_functions.items()}
    else:
        dict_device_kernel_functions = compile_cl_program_device(program_model, context, file)
        callable_kernels = {k: CallableKernelDevice(kernel_model=dict_kernels_program_model[k], compiled=v)
                            for k, v in dict_device_kernel_functions.items()}
    # make callable kernel available in knl model instance
    for knl in program_model.kernels:
        knl.callable_kernel = callable_kernels[knl.name]
    context.add_time_compilation(time.perf_counter_ns() - t_ns_start)
    return ProgramContainer(program_model=program_model,
                            file=file,
                            init=context,
                            callable_kernels=callable_kernels)


def int_safe(val: float):
    if val.is_integer():
        return int(val)
    else:
        raise ValueError(f'val={val} is no integer')


class MutableInteger:
    # e.g. if HashArray is reshaped, a mutable integer is required to preserve reference to original integer
    def __init__(self, val: int):
        self.val = val

    def __eq__(self, other: 'MutableInteger'):
        if isinstance(other, MutableInteger):
            return self.val == other.val
        elif isinstance(other, int):
            raise ValueError('Comparison to integers is not allowed to avoid mistakes that are difficult to debug.')
        else:
            return False

    def __ne__(self, o: 'MutableInteger') -> bool:
        if isinstance(o, MutableInteger):
            return self.val != o.val
        elif isinstance(o, int):
            raise ValueError('Comparison to integers is not allowed to avoid mistakes that are difficult to debug.')
        else:
            return True

    def __str__(self):
        return str(self.val)

    def __copy__(self):
        return MutableInteger(self.val)

    def copy(self):
        return self.__copy__()


class HashArray(Array):
    def __init__(self, *args, **kwargs):
        if isinstance(args[0], TypesClArray.__args__):
            a = args[0]
            super().__init__(a.queue, a.shape, a.dtype, order="C", allocator=a.allocator,
                             data=a.data, offset=a.offset, strides=a.strides, events=a.events)
        else:
            super().__init__(*args, **kwargs)
        self.hash = MutableInteger(0)
        self.update_hash()

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self.update_hash()

    def set(self, ary, queue=None, async_=None, **kwargs):
        res = super().set(ary, queue, async_, **kwargs)
        self.update_hash()
        return res

    def update_hash(self):
        self.hash.val = hash(self.get().tobytes())

    def reshape(self, *shape, **kwargs):
        _hash = self.hash
        _reshaped = super().reshape(*shape, **kwargs)
        _reshaped.hash = _hash
        return _reshaped


class Helpers:
    # helper methods which can be useful in interplay with this framwork
    @staticmethod
    def _camel_to_snake(name):
        name = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
        return re.sub('([a-z0-9])([A-Z])', r'\1_\2', name).lower()

    @staticmethod
    def command_compute_address(n_dim: int) -> str:
        command = '0'
        for i in range(n_dim):
            offset = '1'
            for j in range(i + 1, n_dim):
                offset += '*get_global_size({})'.format(j)
            command += '+get_global_id({})*{}'.format(i, offset)
        return command

    # helpers for using vector types

    @staticmethod
    def get_vec_dtype(dtype_vec: npt.DTypeLike, dtype_scalar: npt.DTypeLike) -> npt.DTypeLike:
        if number_vec_elements_of_cl_type(dtype_vec) == 1:
            return dtype_scalar
        else:
            c_name = '{}{}'.format(c_name_from_dtype(dtype_scalar), number_vec_elements_of_cl_type(dtype_vec))
            return getattr(Types, c_name)

    @staticmethod
    def array_indexing_for_vec_type(array: str, index: str, dtype: npt.DTypeLike):
        """
        https://stackoverflow.com/questions/24746221/using-a-vector-as-an-array-index
        e.g.
        uchar4 keys = (uchar4)(5, 0, 2, 6);
        uint4 results = (uint4)(data[keys.s0], data[keys.s1], data[keys.s2], data[keys.s3]);

        :param dtype:
        :param array:
        :param index:
        :return:
        """
        if number_vec_elements_of_cl_type(dtype) == 1:
            return '{array_name}[{index_name}]'.format(array_name=array, index_name=index)
        else:
            return '({c_type_name})({vector})'.format(c_type_name=c_name_from_dtype(dtype),
                                                      vector=', '.join(
                                                          ['{array_name}[{index_name}.s{i_vec_element}]'.format(
                                                              array_name=array,
                                                              index_name=index,
                                                              i_vec_element=VEC_INDICES[i])
                                                              for i in range(number_vec_elements_of_cl_type(dtype))]))

    @staticmethod
    def command_const_vec_type(param: Union[str, float, int], dtype: npt.DTypeLike) -> str:
        """
        param = 1.5, dtype=ClTypes.float -> 'convert_float(1.5)'
        param = 1.5, dtype=ClTypes.float2 -> '(float2)(convert_float(1.5), convert_float(1.5))

        :param param:
        :param dtype:
        :return:
        """
        if number_vec_elements_of_cl_type(dtype) == 1:
            return 'convert_{}({})'.format(c_name_from_dtype(dtype), param)
        else:
            dtype_c_name = c_name_from_dtype(scalar_type_from_vec_type(dtype))
            return '({})(({}))'.format(c_name_from_dtype(dtype),
                                       ', '.join(['convert_{}({})'.format(dtype_c_name,
                                                                          param)] * get_vec_size(dtype)))

    @staticmethod
    def command_vec_sum(var_name: str, dtype: npt.DTypeLike) -> str:
        """
        Cases:
        float var_name -> return 'var_name'
        float4 var_name -> return 'var_name.s0 + var_name.s1 + var_name.s2 + var_name.s3'
        :param var_name:
        :return:
        """
        if get_vec_size(dtype) == 1:
            return var_name
        else:
            return ' + '.join(
                ['{}.s{}'.format(var_name, VEC_INDICES[i]) for i in range(get_vec_size(dtype))])

    # todo: use splay method of pyopencl library instead
    # from pyopencl.array import splay
    # splay
    @staticmethod
    def _get_local_size_coalesced_last_dim(global_size, desired_wg_size):
        """
        E.g. global_size = (1000, 25) and desired_wg_size=64
        Then a local_size=(2,25) is returned for multiple reasons:

        The work group size must be equal or smaller than the desired work group size.
        We make the last local dimension is large as possible (cannot exceed global size of last dimension).
        If possible the second last dimension is set to a value larger than 1, such that we get close to our desired
        work group size.

        :param global_size:
        :param desired_wg_size:
        :return:
        """

        local_size = [1] * len(global_size)
        for i_dim in range(1, len(global_size) + 1):
            if global_size[-i_dim] * local_size[-i_dim + 1] < desired_wg_size:
                local_size[-i_dim] = global_size[-i_dim]
            else:
                local_size[-i_dim] = np.max([i for i in range(1, desired_wg_size + 1)
                                             if (global_size[-i_dim] / i).is_integer() and
                                             i * local_size[-i_dim + 1] <= desired_wg_size])
        if np.product(local_size) < desired_wg_size:
            pass
            # res = inspect.stack()
            # logging.info(f'Local size {local_size} is suboptimal for desired work group size of {desired_wg_size}. '
            #              f'For best performance increase the global size of the most inner dimension, until it is '
            #              f'divisible by {desired_wg_size}. \n'
            #              f'More information: '
            #              f'https://stackoverflow.com/questions/3957125/questions-about-global-and-local-work-size')
        return tuple(local_size)
        # return None

    @staticmethod
    def get_local_size_coalesced_last_dim(global_size, context: Context):
        """
        If global size is no multiple of the local size, according to following link it should not work.
        https://community.khronos.org/t/opencl-ndrange-global-size-local-size/4167

        However (only for AMD GPU), simple tests have shown that it still works. Therefore this class gives a local size, where the global
        size is not necessarily a multiple.

        :param global_size:
        :param context:
        :return:
        """
        desired_wg_size = 4 * context.devices[0].global_mem_cacheline_size
        return Helpers._get_local_size_coalesced_last_dim(global_size, desired_wg_size)


__author__ = "P.Mohr"
__copyright__ = "26.03.2020, P.Mohr"
__version__ = "1.0"
__email__ = "philipp.mohr@tuhh.de"
__doc__ = """This script includes helpful functions to extended PyOpenCl functionality."""

import logging
from functools import partial
from pathlib import Path
from typing import cast


def typed_partial(cls, *args, **kwargs):
    """
    https://stackoverflow.com/questions/61126905/typing-how-to-consider-class-arguments-wrapped-with-partial
    """
    return cast(cls, partial(cls, *args, **kwargs))


def write_string_to_file(data_str: str, file, b_logging: bool=True):
    """

    :param file:
    :param data_str:
    :param relative_folder_path: e.g. relative_folder_path='./results/
    :return:
    """
    path = Path(file).parent
    if not path.exists():
        Path.mkdir(path, parents=True)
        print('Created directory ' + str(path))
    outfile = open(file, 'w')
    if b_logging:
        logging.info('write: ' + str(file))
    outfile.write(data_str)
    outfile.close()



from pyopencl_extension import Kernel, Global, Scalar, Types, zeros
import numpy as np

ary = zeros((10,), Types.short)

knl = Kernel('some_operation',
             {'ary': Global(ary),  # notice that ary is set as default argument
              'number': Scalar(Types.short)},
             """
                ary[get_global_id(0)] = get_global_id(0) + number;
             """,
             global_size=ary.shape).compile(emulate=True)
knl(number=3)
assert np.allclose(ary.get(), np.arange(10) + 3)



import logging
from pathlib import Path

import numpy as np
from pyopencl import Program as pyopencl_program
from pytest import mark

from pyopencl_extension import emulation, use_existing_file_for_emulation, Local, Scalar, \
    LocalArray, Types, Kernel, Global, Function, Program, Private, zeros, zeros_like, to_device, get_current_queue
from pyopencl_extension.emulation import unparse_c_code_to_python, create_py_file_and_load_module, compute_linear_idx, \
    compute_tuple_idx
from pyopencl_extension.framework import create_cl_files

path_py_cl = Path(__file__).parent.joinpath('py_cl_kernels')

rendered_template = """
#define PI 3.14159265358979323846
#define MUL(x,y) (x*y)
#define ADD(x,y) (x+y)
#define SUB(x,y) (x-y)
#define ABS(x) (fabs(x))
#define RMUL(x,y) (x*y)
#define NEW(x) (x)


#define SOME_CONSTANT 6

typedef short data_t;

float plus_one(__global  short *buffer,
               int idx)
{
    return buffer[idx]+5;
}

__kernel void some_operation(__global  short *buff,
                             short number)
{
    data_t factor = convert_short(1.3);
    buff[get_global_id(0)] = plus_one(buff, get_global_id(0)) + SOME_CONSTANT*factor;
}
"""


def test_debug_kernel():
    buff_cl = zeros((10, 1), Types.short)
    # compute result with opencl
    queue = get_current_queue()
    program = pyopencl_program(queue.context, str(rendered_template)).build()
    some_operation = program.all_kernels()[0]
    # some_operation(init.queue, buff_cl.shape, None,
    #                buff_cl.data, np.dtype(ClTypes.short).type(3))
    # alternative:
    some_operation.set_scalar_arg_dtypes([None, Types.short])
    some_operation(queue, buff_cl.shape, None, buff_cl.data, 3)

    buff_py = zeros((10, 1), Types.short)
    code_py = unparse_c_code_to_python(rendered_template)
    module = create_py_file_and_load_module(code_py, str(path_py_cl.joinpath('debug_component')))

    module.some_operation(buff_py.shape, None,
                          buff_py,
                          np.dtype(Types.short).type(3))

    assert np.all(buff_py.get() == 11)
    assert np.all(buff_py.get() == buff_cl.get())


rendered_template_cplx_float = """
#include <pyopencl-complex.h>
#define TP_ROOT cfloat
#define PI 3.14159265359f

#define MUL cfloat_mul
#define ADD cfloat_add
#define SUB cfloat_sub
#define ABS cfloat_abs
#define RMUL cfloat_rmul
#define NEW cfloat_new

__kernel void some_operation(__global  cfloat_t *buff,
                             cfloat_t number)
{
    buff[get_global_id(0)] = number;
}
"""


def test_debug_kernel_with_complex_numbers():
    buff_cl = zeros((10, 1), Types.cfloat)
    queue = get_current_queue()
    # compute result with opencl
    program = pyopencl_program(queue.context, str(rendered_template_cplx_float)).build()
    some_operation = program.all_kernels()[0]
    # some_operation.set_scalar_arg_dtypes([None, ClTypes.cfloat])
    some_operation(queue, buff_cl.shape, None,
                   buff_cl.data, np.dtype(Types.cfloat).type(3 + 1j))

    buff_py = zeros((10, 1), Types.cfloat)
    code_py = unparse_c_code_to_python(rendered_template_cplx_float)
    module = create_py_file_and_load_module(code_py, str(path_py_cl.joinpath('debug_component_complex_numbers')))

    module.some_operation(buff_py.shape, None,
                          buff_py, np.dtype(Types.cfloat).type(3 + 1j))

    assert np.all(buff_py.get() == buff_cl.get())


rendered_template_cplx_double = """
#if defined(cl_khr_fp64)  // Khronos extension available?
#pragma OPENCL EXTENSION cl_khr_fp64 : enable
#define PYOPENCL_DEFINE_CDOUBLE
#elif defined(cl_amd_fp64)  // AMD extension available?
#pragma OPENCL EXTENSION cl_amd_fp64 : enable
#define PYOPENCL_DEFINE_CDOUBLE
#endif

#include <pyopencl-complex.h>
#define TP_ROOT cfloat
#define PI 3.14159265359f

__kernel void some_operation(__global  cdouble_t *buff,
                             cdouble_t number)
{
    buff[get_global_id(0)] = number;
}
"""


def test_debug_kernel_with_complex_numbers_double():
    cplx_t = Types.cdouble
    buff_cl = zeros((10, 1), cplx_t)
    queue = get_current_queue()
    # compute result with opencl
    program = pyopencl_program(queue.context, str(rendered_template_cplx_double)).build()
    some_operation = program.all_kernels()[0]
    # some_operation.set_scalar_arg_dtypes([None, ClTypes.cfloat])
    some_operation(queue, buff_cl.shape, None,
                   buff_cl.data, np.dtype(cplx_t).type(3 + 1j))

    buff_py = zeros((10, 1), cplx_t)
    code_py = unparse_c_code_to_python(rendered_template_cplx_float)
    module = create_py_file_and_load_module(code_py, str(path_py_cl.joinpath('debug_component_complex_numbers_double')))

    module.some_operation(buff_py.shape, None,
                          buff_py, np.dtype(cplx_t).type(3 + 1j))

    assert np.all(buff_py.get() == buff_cl.get())


def test_debug_try_python_conversion_if_c_compilation_fails():
    pass


# from here on all tests depend on framework
def test_debug_c_code_with_unary_increment_operation_inside_of_array():
    buff_cl = zeros((6, 1), Types.short)
    knl = Kernel('knl',
                 {'buff': Global(buff_cl)},
                 """
        int number = -1;
        number++;
        buff[number++] = 1;
        buff[number] = 2;
        number = 0;
        buff[2+ number--] = 3;
        buff[3+ ++number] = 4;
        buff[5 + --number] = 5;
        int count = 0;
        for(int i=1; i<3; i++){
            count = count + i;
        }        
        buff[5] = count;
    """,
                 global_size=(1,))
    compiled_cl = knl.compile(emulate=False)
    compiled_cl(buff=buff_cl)
    buff_py = zeros((6, 1), Types.short)
    compiled_py = knl.compile(emulate=True, file=Path(__file__).parent.joinpath('py_cl_kernels/knl'))
    compiled_py(buff=buff_py)
    assert np.all(buff_py.get() == buff_cl.get())


def test_access_complex_variable():
    buff = np.array([0.5]).astype(Types.cfloat)
    buff_in = to_device(buff)
    buff_out = zeros_like(buff_in)
    knl = Kernel('knl',
                 {'inp': Global(buff_in),
                  'out': Global(buff_out)},
                 """
        out[get_global_id(0)].real = inp[get_global_id(0)].real; 
    """,
                 global_size=(1,))
    compiled_cl = knl.compile(emulate=False, file=Path(__file__).parent.joinpath('py_cl_kernels/knl'))
    compiled_cl()
    buff_out_py = zeros_like(buff_in)
    compiled_py = knl.compile(emulate=True, file=Path(__file__).parent.joinpath('py_cl_kernels/knl'))
    # out[0] = complex64(inp[0].real+out[0].imag*1j) instead of out[0].real=inp[0].real
    compiled_py(out=buff_out_py)
    assert np.all(buff_out.get() == buff_out_py.get())


def test_debug_kernel_with_barriers():
    buff = np.zeros(shape=(2, 4)).astype(Types.int)
    mem_buffer = to_device(buff)
    knl = Kernel('knl',
                 {'mem_glob': Global(mem_buffer)},
                 """
        __local int mem[2];
        mem[0]=0;
        mem[1]=0;
        mem[get_local_id(1)] = get_local_id(1);
        barrier(CLK_LOCAL_MEM_FENCE);
        mem[get_local_id(1)] = mem[1];
        //barrier(CLK_GLOBAL_MEM_FENCE);
        mem_glob[get_global_id(0)*get_global_size(1)+get_global_id(1)] = mem[get_local_id(1)];
    """,
                 global_size=(2, 4),
                 local_size=(1, 2))
    compiled_cl = knl.compile(emulate=False, file=Path(__file__).parent.joinpath('py_cl_kernels/knl'))
    compiled_cl()
    mem_buffer_py = zeros_like(mem_buffer)
    compiled_py = knl.compile(emulate=True, file=Path(__file__).parent.joinpath('py_cl_kernels/knl'))
    # out[0] = complex64(inp[0].real+out[0].imag*1j) instead of out[0].real=inp[0].real
    compiled_py(mem_glob=mem_buffer_py)
    assert np.all(mem_buffer.get() == mem_buffer_py.get())


def test_number_overflow():
    inp1 = np.array([127, 10, -128]).astype(Types.char)
    inp2 = np.array([127, 10, -128]).astype(Types.char)
    knl = Kernel('knl',
                 {
                     'inp1': inp1,
                     'inp2': inp2,
                     'out1': np.zeros_like(inp1, dtype=Types.char),
                     'out2': np.zeros_like(inp1, dtype=Types.char)},
                 """
        char a = 0;
        a = add_sat(inp1[get_global_id(0)], inp2[get_global_id(0)]);
        char b = 0;
        b = inp1[get_global_id(0)] + inp2[get_global_id(0)];
        out1[get_global_id(0)] = a; 
        out2[get_global_id(0)] = b; 
    """,
                 global_size=inp1.shape)
    knl_cl = knl.compile()
    knl_py = knl.compile(emulate=True)
    knl_cl()
    res_cl = knl_cl.out1.get(), knl_cl.out2.get()
    knl_py()
    res_py = knl_cl.out1.get(), knl_cl.out2.get()
    assert np.all(res_cl[0] == res_py[0]) and np.all(res_cl[1] == res_py[1])


# implement pointer arithmetics with pointer wrapper class for every variable
def test_pointer_arithmetics():
    # Problem: abstract syntax tree does not distinguish if an identifier is a pointer or a variable.
    # E.g. if incrementing the pointer to an array a (a=a+1) in Python this would increment all values in
    # the underlying array. However if
    data = np.array([0, 0]).astype(Types.char)
    knl = Kernel('knl_pointer_arithmetics',
                 {'data': data},
                 """
        char a[5]={0};
        a[0] = a[0] + 1;
        a[1] = 1;
        char* b = a + 1;
        b -= 1; b += 1;
        data[0] = b[0];
        
        char* c = a;
        c += 1;
        a[1] = 3;
        data[1] = c[0];
    """, global_size=data.shape)
    emulation.use_existing_file_for_emulation(False)
    knl_py = knl.compile(emulate=True)
    knl_cl = knl.compile()
    knl_cl()
    res_cl = knl_cl.data.get()
    knl_py()
    res_py = knl_cl.data.get()
    assert np.all(res_cl[0] == res_py[0])


@mark.parametrize('dtype',
                  [Types.char,
                   Types.char4],
                  ids=['scalar type',
                       'vector type'])
def test_pointer_increment(dtype):
    # todo use https://numpy.org/doc/stable/reference/generated/numpy.ndarray.ctypes.html
    data = np.array([0]).astype(dtype)
    func = Function('func',
                    {'data': Private(data.dtype)},
                    """
        return data[0];
    """, returns=data.dtype)
    # Assigning to array pointers does not work in c (e.g. b=a does not compile):
    # https://stackoverflow.com/questions/744536/c-array-declaration-and-assignment
    # Below this can be solved by creating pointers p1 and p2 where their address can be exchange by assignment
    knl = Kernel('knl_pointer_arithmetics',
                 {'data': data},
                 """ private dtype a[5] = {0}; private dtype b[5] = {0};
                     dtype *p1 = a; dtype *p2 = b;
                     a[3] = (dtype)(5);                     
                     p2 = a;
                     data[0] = func(p2+3); """, global_size=data.shape, type_defs={'dtype': dtype})
    prog = Program(functions=[func], kernels=[knl])
    knl_cl = prog.compile().knl_pointer_arithmetics
    knl_cl()
    res_cl = knl_cl.data.get()
    knl_py = prog.compile(emulate=True).knl_pointer_arithmetics
    knl_py()
    res_py = knl_cl.data.get()
    assert np.all(res_cl[0] == res_py[0])


def test_bit_shift():  # todo use https://numpy.org/doc/stable/reference/generated/numpy.ndarray.ctypes.html
    data = np.array([0, 0, 0, 0]).astype(Types.char)
    knl = Kernel('knl_bit_packing',
                 {'data': data},
                 """
        uchar a = 5;
        uchar b = 3;
        uchar c = (a << 4) | b;
        data[0] = (c & 0xF0) >> 4;
        data[1] = c & (0x0F);
    """,
                 global_size=data.shape)
    prog = Program(kernels=[knl])
    knl_cl = prog.compile().knl_bit_packing
    knl_py = prog.compile(emulate=True).knl_bit_packing
    knl_cl()
    get_current_queue().finish()
    res_cl = knl_cl.data.get()
    knl_py()
    res_py = knl_cl.data.get()
    assert np.all(res_cl == res_py)


@mark.parametrize("header", ['for(int i=10; i >=0; i--)',
                             'for(int i=10; i !=0; i--)',
                             'for(int i=10; i >0; i--)',
                             'for(int i=0; i <=10; i++)',
                             'for(int i=0; i !=10; i++)',
                             'for(int i=0; i <10; i++)'])
def test_for_loop(header):
    def eval_code(emulate=False):
        data = to_device(np.array([0]).astype(Types.char))
        knl = Kernel('knl_test_for_loop',
                     {'data': Global(data)},
                     """
            ${header}{
                data[0]+=i;
            }
        """,
                     replacements={'header': header},
                     global_size=data.shape).compile(emulate=emulate)
        knl()
        get_current_queue().finish()
        res = knl.data.get()
        return res

    assert np.all(eval_code(False) == eval_code(True))


def test_vector_types():  # todo use https://numpy.org/doc/stable/reference/generated/numpy.ndarray.ctypes.html
    data = np.zeros((10,)).astype(Types.char2)
    knl = Kernel('knl_vector_types',
                 {'data': data},
                 """
        char2 a = (char2)(4,2);
        char2 b = (char2)(1,2);
        data[0] = a;
        data[1] = b;
        data[2] = a + b;
        data[3] = a * b;
        data[4] = a - b;
        data[5] = a / b;
    """,
                 global_size=data.shape)
    knl_cl = knl.compile()
    knl_py = knl.compile(emulate=True)
    knl_cl()
    get_current_queue().finish()
    res_cl = knl_cl.data.get()
    knl_py()
    res_py = knl_py.data.get()
    assert np.all(res_cl == res_py)


def test_nested_local_barrier_inside_function():
    func_nested = Function('nested_func',
                           {
                               'ary': Global(Types.int),
                               'shared': Local(Types.int)},
                           """
               shared[get_global_id(0)] = ary[get_global_id(0)];
               barrier(CLK_LOCAL_MEM_FENCE);
               return shared[(get_global_id(0)+1)%2] ;
               """,
                           returns=Types.int)
    func_parent = Function('parent',
                           func_nested.args,
                           """
               return nested_func(ary, shared);
               """,
                           returns=Types.int)

    ary = to_device((ary_np := np.array([1, 2]).astype(Types.int)))
    use_existing_file_for_emulation(False)
    knl = Kernel('some_knl',
                 {
                     'ary': Global(ary),
                 },
                 """
                __local int shared[2];
                ary[get_global_id(0)] = parent(ary, shared);
                   """, global_size=ary.shape)
    prog = Program([func_nested, func_parent], [knl])
    prog_py = prog.compile(emulate=True)
    prog_cl = prog.compile(emulate=False)
    prog_py.some_knl()

    ary_py = ary.get()
    ary.set(ary_np)
    prog_cl.some_knl()
    ary_cl = ary.get()
    get_current_queue().finish()
    assert np.allclose(ary_py, np.array([2, 1]))
    assert np.allclose(ary_py, ary_cl)


def test_macro_with_arguments():
    defines = {'FUNC(a,b,c)': '{ int tmp = c(a-b); a += b + tmp; }'}  # this is a macro with arguments
    ary = zeros((2,), Types.int)
    func_add_two = Function('add_two',
                            {'a': Scalar(Types.int)},
                            'return a + 2;', returns=Types.int)
    knl = Kernel('knl_macro_func',
                 {'ary': Global(ary)},
                 """
               int a = 1;
               int b = 2;
               FUNC(a, b, add_two)
               ary[get_global_id(0)] = a;
               """,
                 defines=defines,
                 global_size=ary.shape)
    Program([func_add_two], [knl]).compile().knl_macro_func()
    assert np.allclose(ary.get(), np.array([4, 4]).astype(ary.dtype))


# this test acts as an integration test, testing multiple different c operations:
# -ternary operator (e.g. x=cond ? a:b;)
# - ...
# Testing with one large integration test reduces test time compared to multiple individual tests.
def test_different_c_operations_at_once():
    ary = zeros((2,), Types.int)
    knl = Kernel('knl_multiple_c_operations',
                 {'ary': Global(ary)},
                 """int a = 1;
                    int b = 2;
                    dtype val; // test variable definition without assignment
                    dtype *ptr1; // test pointer definition 
                    global dtype *ptr2; // test global pointer definition 
                    ary[get_global_id(0)] = a>get_global_id(0) ? a : b;
                 """, global_size=ary.shape, type_defs={'dtype': ary.dtype}).compile(emulate=True)
    knl()
    assert np.allclose(ary.get(), np.array([1, 2]).astype(ary.dtype))


# https://community.khronos.org/t/allocate-array-in-a-kernel-of-length-known-only-at-runtime/3655/3
# http://developer.amd.com/wordpress/media/2013/12/AMD_OpenCL_Programming_User_Guide2.pdf
# http://developer.amd.com/wordpress/media/2013/01/Introduction_to_OpenCL_Programming-Training_Guide-201005.pdf
@mark.parametrize(argnames='data_t', argvalues=[Types.short, Types.double])
def test_local_memory_as_kernel_argument(data_t):
    def run(emulate=False):
        ary = to_device(np.ones(10).astype(data_t))
        local_mem = LocalArray(dtype=data_t, shape=5)  # 5 is to to test that local array argument is changed
        knl = Kernel('knl_local_arg',
                     {'ary': Global(ary), 'local_mem': local_mem},
                     """
               int offset = get_group_id(0)*get_local_size(0);
               for(int i=0; i<5; i++) local_mem[i] = ary[offset + i];
               barrier(CLK_LOCAL_MEM_FENCE);
               data_t sum = (data_t)(0);
               for(int i=0; i<5; i++) sum+=local_mem[i];
               ary[get_global_id(0)] = sum;
                     """,
                     type_defs={'data_t': data_t},
                     global_size=ary.shape,
                     local_size=(5,))
        local_mem = LocalArray(dtype=data_t, shape=5)
        knl.compile(emulate=emulate)(local_mem=local_mem)
        return ary.get()

    ary_cl = run(emulate=False)
    ary_py = run(emulate=True)
    assert np.allclose(ary_cl, ary_py) and np.allclose(ary_cl, 5 * np.ones(10).astype(ary_py.dtype))


# constants in global scope cannot be set from host.
# https://stackoverflow.com/questions/7140820/opencl-initializing-program-scope-variables-from-the-host
# def test_program_scope_variable(thread):
#     pass


def test_barrier_global_local_mem_fence():
    pass


@mark.parametrize(['name', 'dtype'], [('fmax', Types.float),
                                      ('abs_diff', Types.int),
                                      ('abs_diff', Types.int),
                                      ('max', Types.int),
                                      ])
def test_two_input_integer_functions(name, dtype):
    a_cl = to_device(np.ones((10,), dtype))
    a_emulation = to_device(np.ones((10,), dtype))
    knl = Kernel(f'knl_{name}',
                 {'a': Global(a_cl), 'num': Scalar(dtype(0))},
                 f'a[get_global_id(0)]={name}(a[get_global_id(0)], num);', global_size=a_cl.shape)
    knl.compile()(a=a_cl)
    knl.compile(emulate=True)(a=a_emulation)
    assert np.all(a_cl.get() == a_emulation.get())



import pyopencl_extension as cl


def test_kernel_compile_on_call_and_abbreviations_work_item_builtin_fncs():
    a = cl.zeros((10,), dtype=cl.int)
    knl = cl.Kernel('my_knl', {'a': a},
                    """a[gid0]+=1.0; a[gid0] = a[gid0];""",
                    global_size=a.shape)
    knl_py = knl.compile(emulate=True)
    # dfs = {}
    # knl = cl.Kernel('my_knl', {'a': a}, 'a[get_global_id(0)]+=1.0;', global_size=a.shape, defines=dfs).compile(emulate=True)
    knl()  # compile on execution if not compiled yet
    # knl(emulate=True)
    assert a.get()[0] == 1.0
    knl_py()
    assert a.get()[0] == 2.0


@mark.parametrize('idx_tuple,dimensions, idx_lin_ref', [((1, 2, 0), (2, 4, 2), 12),
                                                        ((1, 2), (2, 4), 6),
                                                        ((1,), (2,), 1)])
def test_compute_tuple_from_idx_linear(idx_tuple, dimensions, idx_lin_ref):
    idx_lin = compute_linear_idx(idx_tuple, dimensions)
    assert idx_lin == idx_lin_ref
    idx_tuple2 = compute_tuple_idx(idx_lin, dimensions)
    assert idx_tuple == idx_tuple2


def test_kernel_pointer_decl():
    a = cl.zeros((10,), dtype=cl.int)
    knl = cl.Kernel('my_knl', {'a': a},
                    """
                    global int* c_glob = a;
                    private int ary[5];
                    int* c;
                    int* b=ary;
                    c = b; // swap pointers
                    c[1] = 5;
                    a[1] = ary[1];
                    c_glob[2] = 10;""",
                    global_size=a.shape)
    knl_py = knl.compile(emulate=True)
    knl_py()
    assert a.get()[1] == 5   # verify private pointer behavior
    assert a.get()[2] == 10  # verify global pointer behavior global int* c_glob = a;
    a2 = cl.zeros((10,), dtype=cl.int)
    knl(a=a2)
    assert a2.get()[1] == 5
    assert a2.get()[2] == 10



import re as re
from typing import Union, Callable

import numpy as np

__author__ = "P.Mohr"
__copyright__ = "26.03.2020, P.Mohr"
__version__ = "1.0"
__email__ = "philipp.mohr@tuhh.de"
__doc__ = """This script includes helpful functions to extended PyOpenCl functionality."""

from pyopencl_extension.types.auto_gen.cl_types import ClTypesVector
from pyopencl_extension.types.auto_gen.cl_types import _ClTypes

np_to_c_type_name = {
    'float16': 'half',
    'float32': 'float',
    'float64': 'double',
    'int64': 'long',
    'uint64': 'ulong',
    'int32': 'int',
    'uint32': 'uint',
    'int16': 'short',
    'uint16': 'ushort',
    'int8': 'char',
    'uint8': 'uchar',
    'complex64': 'cfloat_t',
    'complex128': 'cdouble_t',
    'void': 'void'
}

c_to_np_type_name = {v: k for k, v in np_to_c_type_name.items()}


def np_to_c_type_name_catch(np_name):
    try:
        return np_to_c_type_name[np_name]
    except:
        return np_name


def c_to_np_type_name_catch(c_name):
    try:
        return c_to_np_type_name[c_name]
    except:
        return c_name


class Types(_ClTypes):
    cfloat: Union[np.dtype, Callable] = np.dtype(np.complex64)
    cdouble: Union[np.dtype, Callable] = np.dtype(np.complex128)


VEC_INDICES = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 'A', 'B', 'C', 'D', 'E', 'F']
VEC_INDICES_XYZW = ['x', 'y', 'z', 'w']
VEC_INDICES_SPECIAL = ['lo', 'hi', 'odd', 'even']
# for details see https://www.khronos.org/files/opencl-1-2-quick-reference-card.pdf
VEC_ADDRESSING = {
    'lo': {'2': [0],
           '3': [0, 1],  # When using .lo or .hi with a 3-component vector, the .w component is undefined
           '4': [0, 1],
           '8': [0, 1, 2, 3],
           '16': [0, 1, 2, 3, 4, 5, 6, 7]},
    'hi': {'2': [1],
           '3': [2, 1],
           '4': [2, 3],
           '8': [4, 5, 6, 7],
           '16': [8, 9, 'A', 'B', 'C', 'D', 'E', 'F']},
    'odd': {'2': [1],
            '3': [1, 3],
            '4': [1, 3],
            '8': [1, 3, 5, 7],
            '16': [1, 3, 5, 7, 9, 'B', 'D', 'F']},
    'even': {'2': [0],
             '3': [0, 2],
             '4': [0, 2],
             '8': [0, 2, 4, 6],
             '16': [0, 2, 4, 6, 8, 'A', 'C', 'E']}
}


def get_vec_size(dtype: np.dtype) -> int:
    # if dtype is scalar type return 1
    if not hasattr(dtype, 'fields') or dtype.fields is None:
        return 1
    else:
        return len(dtype.names)


def c_name_from_dtype(dtype: np.dtype) -> str:
    """
    :param dtype: A numpy data type class
    :return: The C-Type corresponding to the numpy dtype
    """
    if not hasattr(dtype, 'fields') or dtype.fields is None:  # scalar type
        try:
            return np_to_c_type_name[dtype.name]
        except:
            return np_to_c_type_name[dtype.__name__]
    else:  # vector type
        n_dim = get_vec_size(dtype)
        name = np_to_c_type_name[dict(dtype.fields)[list(dtype.fields)[0]][0].name]
        return '{}{}'.format(name, n_dim)


def dtype_from_c_name(c_name: str):
    return getattr(Types, c_name)


def dtype_to_c_name(dtype):
    return np_to_c_type_name_catch(dict(dtype.fields)['s0'][0].name) + str(len(dtype.names))


def get_unsigned_dtype(dtype):
    return dtype_from_c_name('u' + c_name_from_dtype(dtype))


def scalar_type_from_vec_type(dtype: np.dtype) -> np.dtype:
    if number_vec_elements_of_cl_type(dtype) == 1:
        return dtype
    else:
        c_vec_name = c_name_from_dtype(dtype)
        return getattr(Types, re.search(r'([a-z]+)([\d]+)', c_vec_name).group(1))


def number_vec_elements_of_cl_type(dtype: np.dtype) -> int:
    if not hasattr(dtype, 'names'):
        return 1
    elif dtype.names is None:
        return 1
    else:
        return len(dtype.names)
    # if 'double' in c_name_from_dtype(dtype):
    #     return int(dtype.itemsize / 8)
    # else:
    #     return dtype.itemsize


def match_vec_size(desired_type: np.dtype, match_vec_type: np.dtype) -> np.dtype:
    """

    :param desired_type:
    :param match_vec_type:
    :return: e.g. desired_type=ClTypes.float, match_vec_type=ClTypes.int4 -> return ClType.float4
    """
    desired_vec_size = get_vec_size(match_vec_type)
    desired_type_scalar = scalar_type_from_vec_type(desired_type)
    if desired_vec_size == 1:
        return desired_type_scalar
    else:
        c_name = '{}{}'.format(c_name_from_dtype(desired_type_scalar), desired_vec_size)
        return dtype_from_c_name(c_name)


def b_is_whole_number(t: np.dtype) -> bool:
    return scalar_type_from_vec_type(t) not in [Types.float, Types.double]


def match_integer_type_for_select(dtype: np.dtype):
    """
    When using overloadable select(a,b,c), the correct function is found by the types.
    E.g. for certain types like half or double this type can be either short or ushort.
    We must provide the type with a cast in order to avoid compilation error.
    select(half16 a, half16 b, short16 c);
    select(half16 a, half16 b, ushort16 c);
    match_integer_type_for_select(dtype=ClTypes.half16)->Return short16
    """
    dtype_scalar = scalar_type_from_vec_type(dtype)
    name = c_name_from_dtype(dtype_scalar)
    if name.startswith('u'):
        name = name[1:]
    conversions = {
        'char': 'char',
        'short': 'short',
        'int': 'int',
        'long': 'long',
        'float': 'int',
        'half': 'short',
        'double': 'long',
    }
    c_name_condition_type = conversions[name]
    # unsigned variant sometimes does not work, signed always works
    # if b_prefer_unsigned:
    #     c_name_condition_type ='u{}'.format(c_name_condition_type)
    dtype_condition = match_vec_size(dtype_from_c_name(c_name_condition_type), dtype)
    return dtype_condition


def scalar_to_vector_type_array(ary: np.ndarray, vec_size: int = 1):
    pass


def is_signed_integer_type(dtype: np.dtype) -> bool:
    scalar_dtype = scalar_type_from_vec_type(dtype)
    if scalar_dtype in [Types.char, Types.short, Types.int, Types.long]:
        return True
    else:
        return False


def is_complex_type(dtype: np.dtype):
    if dtype in [Types.cfloat, Types.cdouble]:
        return True
    else:
        return False


def is_vector_type(dtype: np.dtype):
    return dtype in list(ClTypesVector().__dict__.values())


def defines_generic_operations(cl_code, generic_type: np.dtype):
    if is_complex_type(generic_type):
        preample_cplx_operations = """
        #define MUL c${cplx_type}_mul
        #define ADD c${cplx_type}_add
        #define SUB c${cplx_type}_sub
        #define ABS c${cplx_type}_abs
        #define RMUL c${cplx_type}_rmul
        #define NEW c${cplx_type}_new
        #define CONJ c${cplx_type}_conj
        """.replace('${cplx_type}', 'float' if generic_type is Types.cfloat else 'double')
    else:
        preample_real_operations = """
        #define MUL c${cplx_type}_mul
        #define ADD c${cplx_type}_add
        #define SUB c${cplx_type}_sub
        #define ABS c${cplx_type}_abs
        #define RMUL c${cplx_type}_rmul
        #define NEW c${cplx_type}_new
        #define CONJ c${cplx_type}_conj
        """.replace('${cplx_type}', 'float' if generic_type is Types.cfloat else 'double')

    return cl_code



from pyopencl_extension import Helpers, Program, Kernel, Global, Scalar, Array, empty
from pyopencl_extension import Types

__author__ = "P.Mohr"
__copyright__ = "26.03.2020, P.Mohr"
__version__ = "1.0"
__email__ = "philipp.mohr@tuhh.de"
__doc__ = """This script includes helpful functions to extended PyOpenCl functionality."""


class SumAlongAxis:
    """
    Responsibility:
    Calculates the sum over all elements of an OpenCl array along an axis.
    """

    def _command_compute_address_in(self):
        """
        e.g. in_buffer = zeros(shape=(100,10)), axis=0
            out_buffer = zeros(shape=(10,))
        //from for loop in kernel: i = glob_id_axis;
        sum + = in_buffer[i* get_global_size(0) + get_global_id(0)];

        e.g. in_buffer = zeros(shape=(100,10)), axis=1
            out_buffer = zeros(shape=(100,))
        //from for loop in kernel: i = glob_id_axis;
        sum + = in_buffer[get_global_id(0)* size_axis + i];

        e.g. in_buffer = zeros(shape=(1000, 100,10)), axis=1
            out_buffer = zeros(shape=(1000, 10))
        //from for loop in kernel: i = glob_id_axis;
        sum + = in_buffer[get_global_id(0)* size_axis * get_global_size(1) + i * get_global_size(1) + get_global_id(1)];
        """
        command = '0'
        subtract_axis = 0
        for axis_in_buffer in range(self.in_buffer.ndim):
            offset = '1'
            # get_global_id(i) * offset
            # classic way: offset = get_global_size(i+1)*...*get_global_size(ndim-1)
            if axis_in_buffer == self.axis:
                subtract_axis = 1  # subtract to account for smaller global_size than in_buffer size.
            for axis_out_buffer in range(axis_in_buffer - subtract_axis + 1, self.out_buffer.ndim):
                offset += '*get_global_size({})'.format(axis_out_buffer)
            # Now insert size axis to account for larger ndim of in_buffer compared to out_buffer
            if self.axis > axis_in_buffer:
                offset += '*{size_axis}'.format(size_axis=self.in_buffer.shape[self.axis])

            if axis_in_buffer == self.axis:
                command += '+i*{}'.format(offset)
            else:
                axis_out_buffer = axis_in_buffer - subtract_axis
                command += '+get_global_id({})*{}'.format(axis_out_buffer, offset)
        return command

    def _get_cl_program(self) -> Program:
        knl = Kernel(name='sum_along_axis',
                     args={'in_buffer': Global(self.in_buffer),
                           'axis': Scalar(Types.int(self.axis)),
                           'out_buffer': Global(self.out_buffer)},
                     body=["""
                 buff_t sum = (buff_t) 0;
                 for(int i=0; i<${size_input_axis}; i++){// i == glob_id_axis
                    sum+=in_buffer[${addr_in}];
                 }                 
                 out_buffer[${addr}] = sum;
                 """],
                     replacements={'size_input_axis': self.in_buffer.shape[self.axis],
                                   'addr': Helpers.command_compute_address(self.out_buffer.ndim),
                                   'addr_in': self._command_compute_address_in()},
                     global_size=self.out_buffer.shape)
        type_defs = {'buff_t': self.in_buffer.dtype}
        return Program(type_defs=type_defs, kernels=[knl])

    def __init__(self, in_buffer: Array, axis: int = 0, out_buffer: Array = None):
        self.in_buffer = in_buffer
        self.axis = axis
        if self.in_buffer.ndim == 1:
            shape_out = (1,)
        else:
            shape_out = tuple([size for i_axis, size in enumerate(self.in_buffer.shape) if i_axis != self.axis])
        if out_buffer is None:
            self.out_buffer = empty(shape_out, self.in_buffer.dtype)
        else:
            if out_buffer.shape != shape_out:
                raise ValueError('out buffer shape does not match required shape')
            self.out_buffer = out_buffer
        self.program = self._get_cl_program().compile()

    def __call__(self, *args, **kwargs):
        self.program.sum_along_axis()
        return self.out_buffer



import numpy as np

from pyopencl_extension import CopyArrayRegion, Slice, to_device


def test_copy_array_region_on_device():
    ary_np = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.complex64)
    in_cl = to_device(ary_np)
    copy_region = CopyArrayRegion(in_cl, region_in=Slice[0:1, 1:3])

    out = copy_region()

    out_np = in_cl.get()[0:1, 1:3]

    assert np.all(out.get() == out_np)


def test_copy_array_region_on_device_region_none():
    ary_np = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.complex64)
    in_cl = to_device(ary_np)
    copy_region = CopyArrayRegion(in_cl, region_in=None)

    out = copy_region()

    out_np = in_cl.get()

    assert np.all(out.get() == out_np)


def test_copy_array_region_on_device_between_buffers():
    ary_np = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.complex64)
    in_cl = to_device(ary_np)
    out_np = np.zeros(shape=(4, 4), dtype=in_cl)
    out_cl = to_device(out_np)
    copy_region = CopyArrayRegion(in_buffer=in_cl,
                                  region_in=Slice[:1, 1:3],
                                  out_buffer=out_cl,
                                  region_out=Slice[1:2, 2:4])
    out = copy_region()
    out_np[1:2, 2:4] = in_cl.get()[0:1, 1:3]
    assert np.all(out.get() == out_np)


def test_copy_array_region_on_device_negative_indexing():
    ary_np = np.ones(shape=(5, 10), dtype=np.int32)
    in_cl = to_device(ary_np)
    copy_region = CopyArrayRegion(in_buffer=in_cl,
                                  region_in=Slice[:, 2:-1])
    out = copy_region()
    out_np = in_cl.get()[:, 2:-1]
    assert np.all(out.get() == out_np)


def test_copy_array_region_on_device_given_axis_index():
    ary_np = np.ones(shape=(5, 10), dtype=np.int32)
    in_cl = to_device(ary_np)
    copy_region = CopyArrayRegion(in_buffer=in_cl,
                                  region_in=Slice[2:-1, :])
    out = copy_region()
    out_np = in_cl.get()[2:-1, :]
    assert np.all(out.get() == out_np)

# todo
# def test_step_width_larger_one(thread):
#     ary_np = np.array([np.arange(10, dtype=np.int32)] * 2)
#     in_cl = cl.array.to_device(thread.queue, ary_np)
#     out_cl = cl.array.to_device(thread.queue, ary_np)
#     copy_region = CopyArrayRegion(in_buffer=in_cl,
#                                   region_in=Slice[:, ::2],
#                                   out_buffer=out_cl[:, 1::2])
#     out = copy_region()
#     out_np = in_cl.get()[:, ::2]
#     assert np.all(out.get() == out_np)



from pyopencl_extension.modifications_pyopencl import cltypes
import numpy as np
char2=cltypes.char2
char4=cltypes.char4
char8=cltypes.char8
char16=cltypes.char16
short2=cltypes.short2
short4=cltypes.short4
short8=cltypes.short8
short16=cltypes.short16
int2=cltypes.int2
int4=cltypes.int4
int8=cltypes.int8
int16=cltypes.int16
long2=cltypes.long2
long4=cltypes.long4
long8=cltypes.long8
long16=cltypes.long16
uchar2=cltypes.uchar2
uchar4=cltypes.uchar4
uchar8=cltypes.uchar8
uchar16=cltypes.uchar16
ushort2=cltypes.ushort2
ushort4=cltypes.ushort4
ushort8=cltypes.ushort8
ushort16=cltypes.ushort16
uint2=cltypes.uint2
uint4=cltypes.uint4
uint8=cltypes.uint8
uint16=cltypes.uint16
ulong2=cltypes.ulong2
ulong4=cltypes.ulong4
ulong8=cltypes.ulong8
ulong16=cltypes.ulong16
half2=cltypes.half2
half4=cltypes.half4
half8=cltypes.half8
half16=cltypes.half16
float2=cltypes.float2
float4=cltypes.float4
float8=cltypes.float8
float16=cltypes.float16
double2=cltypes.double2
double4=cltypes.double4
double8=cltypes.double8
double16=cltypes.double16
char=cltypes.char
short=cltypes.short
int=cltypes.int
long=cltypes.long
uchar=cltypes.uchar
ushort=cltypes.ushort
uint=cltypes.uint
ulong=cltypes.ulong
half=cltypes.half
float=cltypes.float
double=cltypes.double

cfloat=np.dtype(np.complex64)
cdouble=np.dtype(np.complex128)
                         


from pyopencl_extension.modifications_pyopencl import cltypes
from pyopencl_extension.types.type_handler import *
int_ = int
float_=float
char=TypeHandlerScalar(cltypes.char)
short=TypeHandlerScalar(cltypes.short)
int=TypeHandlerScalar(cltypes.int)
long=TypeHandlerScalar(cltypes.long)
uchar=TypeHandlerScalar(cltypes.uchar)
ushort=TypeHandlerScalar(cltypes.ushort)
uint=TypeHandlerScalar(cltypes.uint)
ulong=TypeHandlerScalar(cltypes.ulong)
half=TypeHandlerScalar(cltypes.half)
float=TypeHandlerScalar(cltypes.float)
double=TypeHandlerScalar(cltypes.double)
char2=TypeHandlerVec('char2')
char4=TypeHandlerVec('char4')
char8=TypeHandlerVec('char8')
char16=TypeHandlerVec('char16')
short2=TypeHandlerVec('short2')
short4=TypeHandlerVec('short4')
short8=TypeHandlerVec('short8')
short16=TypeHandlerVec('short16')
int2=TypeHandlerVec('int2')
int4=TypeHandlerVec('int4')
int8=TypeHandlerVec('int8')
int16=TypeHandlerVec('int16')
long2=TypeHandlerVec('long2')
long4=TypeHandlerVec('long4')
long8=TypeHandlerVec('long8')
long16=TypeHandlerVec('long16')
uchar2=TypeHandlerVec('uchar2')
uchar4=TypeHandlerVec('uchar4')
uchar8=TypeHandlerVec('uchar8')
uchar16=TypeHandlerVec('uchar16')
ushort2=TypeHandlerVec('ushort2')
ushort4=TypeHandlerVec('ushort4')
ushort8=TypeHandlerVec('ushort8')
ushort16=TypeHandlerVec('ushort16')
uint2=TypeHandlerVec('uint2')
uint4=TypeHandlerVec('uint4')
uint8=TypeHandlerVec('uint8')
uint16=TypeHandlerVec('uint16')
ulong2=TypeHandlerVec('ulong2')
ulong4=TypeHandlerVec('ulong4')
ulong8=TypeHandlerVec('ulong8')
ulong16=TypeHandlerVec('ulong16')
half2=TypeHandlerVec('half2')
half4=TypeHandlerVec('half4')
half8=TypeHandlerVec('half8')
half16=TypeHandlerVec('half16')
float2=TypeHandlerVec('float2')
float4=TypeHandlerVec('float4')
float8=TypeHandlerVec('float8')
float16=TypeHandlerVec('float16')
double2=TypeHandlerVec('double2')
double4=TypeHandlerVec('double4')
double8=TypeHandlerVec('double8')
double16=TypeHandlerVec('double16')


import numpy as np
import pyopencl as cl


def get_devices():
    """
    On a computer often multiple chips exist to execute OpenCl code, like Intel, AMD or Nvidia GPUs or FPGAs.
    This function return a list of all available devices.
    """
    platforms = cl.get_platforms()
    devices = [d for p in platforms for d in p.get_devices(device_type=cl.device_type.GPU)]
    # devices = [platform[device[0]].get_devices(device_type=cl.device_type.GPU)[device[1]]][0]
    return devices


class Context(cl.Context):
    def __init__(self, devices, properties=None, dev_type=None):
        if len(devices) > 1:
            raise NotImplementedError('In current implementation only a single device per context is supported. '
                                      'Please request feature if required.')
        super().__init__(devices, properties, dev_type)
        self.time_compilation_ns = 0

    @property
    def device_id(self):
        matches = np.where(np.array([d.int_ptr for d in get_devices()]) == self.devices[0].int_ptr)
        if not len(matches) == 1:
            raise ValueError('int_ptr is expected to be unique identifier for device. '
                             'Change current implementation to fix issue')
        return matches[0][0]

    def add_time_compilation(self, t_ns):
        self.time_compilation_ns += t_ns



import math as math
from abc import abstractmethod

from pyopencl_extension.types.auto_gen.types_for_emulation import *
from pyopencl_extension.types.utilities_np_cl import c_to_np_type_name_catch, is_vector_type, get_unsigned_dtype

cfloat_t = np.dtype('complex64').type
cdouble_t = np.dtype('complex128').type
void = np.dtype('void').type


class CPointerBase:
    @property
    @abstractmethod
    def np(self) -> np.ndarray:
        pass

    def __str__(self):
        return str(self.np)

    def __init__(self, size, type_c, mem=None, pos=0):
        self.type_c = type_c
        self.size = size
        self.memory = self._init_memory() if mem is None else mem
        self._validate_pos(pos)
        self.pos = pos

    @classmethod
    def from_np(cls, ary):
        return cls(size=ary.size, type_c=ary.dtype, mem=ary)

    def view(self, type_c):
        mem = self.memory.view(type_c)
        return self.__class__(type_c=type_c, size=mem.shape[0], mem=mem, pos=self.pos)

    def _validate_pos(self, pos):
        if not 0 <= pos < self.size:
            raise MemoryError('Out of memory bounds positioning of pointer')
        else:
            return pos

    def fill(self, val):
        self.memory.fill(val)

    def __add__(self, other):
        return self.__class__(size=self.size, type_c=self.type_c, mem=self.memory, pos=self.pos + other)

    def __sub__(self, other):
        return self.__class__(size=self.size, type_c=self.type_c, mem=self.memory, pos=self.pos - other)

    def __getitem__(self, item):
        return self.memory[self.pos + item]

    @abstractmethod
    def __setitem__(self, instance, value):
        pass

    def _init_memory(self):
        mem = np.empty((self.size,), dtype=self.type_c)
        mem[:] = 99999  # arbitrary number to indicate that element of array contains intial value
        return mem


class CPointer(CPointerBase):
    @property
    def np(self) -> np.ndarray:
        return self.memory[self.pos:]

    def __setitem__(self, instance, value):
        self.memory[self.pos + instance] = value


class CPointerVec(CPointerBase):
    @property
    def np(self):
        vector_element_dtype = dict(self.memory.dtype.fields)['s0'][0]
        vector_len = len(self.memory.dtype.descr)
        res = np.array(self.memory.view(vector_element_dtype)).reshape(self.memory.shape[0], vector_len)
        return res[self.pos:, :]

    def __setitem__(self, instance, value):
        vec_size = len(self.memory.dtype.descr)
        # get reference
        element = self.memory[instance]
        # element =self[instance] this returns a copy of the the element in memory and therefore an assignment has no effect
        # on original
        for i in range(vec_size):
            field = f's{i}'
            element[field] = value.val[field]
            # self[instance].val[field] = value.val[field]

    def __getitem__(self, item):
        if isinstance(item, slice):
            return self.memory[self.pos]
        else:
            res = self.memory[item + self.pos]
            return VecVal(res.copy())


sign = lambda x: np.sign(x)
fabs = lambda x: np.abs(x)
ceil = lambda x: np.ceil(x)
log2 = lambda x: np.log2(x)
sqrt = lambda x: x ** 0.5
exp = lambda x: np.exp(x)
log = lambda x: np.log(x)
pow = lambda x, y: x ** y
barrier = lambda x: x
CLK_GLOBAL_MEM_FENCE = None
select = lambda a, b, c: b if c else a
any = lambda x: np.any(x)
INFINITY = np.inf


def init_array(size, type_c):
    if isinstance(type_c, TypeHandlerScalar):
        return CPointer(size, type_c)
    elif isinstance(type_c, TypeHandlerVec):
        return CPointerVec(size, type_c)
    else:
        raise ValueError(f'Array initialized with {str(type_c)} not supported')


# init_array = lambda size, type_c: np.empty((size,), dtype=c_to_np_type_name_catch(type_c))


def add_sat(x, y):
    res = np.add(x, y, dtype=int)
    if res > np.iinfo(x.dtype).max:
        return x.dtype.type(np.iinfo(x.dtype).max)
    elif res < np.iinfo(x.dtype).min:
        return x.dtype.type(np.iinfo(x.dtype).min)
    else:
        return res


def abs(x):
    return get_unsigned_dtype(x.dtype.type)(np.abs(x))


def abs_diff(x, y):
    return get_unsigned_dtype(x.dtype.type)(np.abs(x - y))


def set_real(ary, idx, value):
    ary[idx] = value + 1j * ary[idx].imag


def set_imag(ary, idx, value):
    ary[idx] = ary[idx].real + 1j * value


# https://www.askpython.com/python/python-modulo-operator-math-fmod
def c_modulo(dividend, divisor):
    if type(dividend).__module__ == 'numpy':
        return dividend.dtype.type(math.fmod(dividend, divisor))
    elif type(dividend) == int or type(dividend) == int_:
        return int(math.fmod(dividend, divisor))
    else:
        raise NotImplementedError(f'modulo division not implemented for divendend of type {type(dividend)}')


def max(x, y):
    return np.max([x, y])


def min(x, y):
    return np.min([x, y])


def fmax(x, y):
    return np.fmax(x, y)


def clamp(x, minval, maxval):
    return min(max(x, minval), maxval)
